# SPDX-FileCopyrightText: 2021-2022 Rene Luria <rene.luria@infomaniak.com>
# SPDX-License-Identifier: Apache-2.0
"""
This plugin enables usage of Infomaniak public API to complete``dns-01`` challenges.
"""
